from taktile_auth.entities.permission import Permission, PermissionDefinition
from taktile_auth.entities.resource import Resource, ResourceDefinition
from taktile_auth.entities.role import Role, RoleDefinition

__all__ = [
    "Permission",
    "PermissionDefinition",
    "Resource",
    "ResourceDefinition",
    "Role",
    "RoleDefinition",
]



import typing as t

from pydantic import UUID4, BaseModel, parse_obj_as

from taktile_auth.entities import Permission, Role
from taktile_auth.exceptions import InsufficientRightsException
from taktile_auth.parser import RESOURCES, ROLES
from taktile_auth.parser.utils import parse_permission


class TaktileIdToken(BaseModel):
    iss: str
    sub: str
    aud: str
    exp: int
    iat: int
    roles: t.List[str]
    actor_name: str | None = None

    @classmethod
    def build_role(cls, role_signature: str) -> Role:
        role_name, _, role_args_str = role_signature.partition("/")
        role_args = role_args_str.split(",") if role_args_str else []
        role_def = ROLES[role_name]
        return ROLES[role_name].build(**dict(zip(role_def.args, role_args)))

    @classmethod
    def build_permission(cls, permission_str: str) -> Permission:
        permission = parse_permission(permission_str)
        resource_vals = {}

        if permission["resource_name"] not in RESOURCES:
            # by default we deny access to non-existing resources
            raise InsufficientRightsException

        # First, get the resource arguments from the path
        resource_args = permission.get("resource_args", [])
        for i, key in enumerate(
            RESOURCES[permission["resource_name"]].args.keys()
        ):
            if i < len(resource_args):
                resource_vals[key] = resource_args[i]

        # Then, update with any explicit arguments from the permission string
        resource_vals.update(permission.get("args", {}))

        return Permission(
            actions=set(permission["actions"]),
            resource=RESOURCES[permission["resource_name"]].get_resource()(
                **resource_vals
            ),
            resource_name=permission["resource_name"],
        )

    def _is_allowed(
        self, permission_str: str, roles: t.Generator[Role, None, None]
    ) -> None:
        permission = self.build_permission(permission_str)
        if not any(permission in role for role in roles):
            raise InsufficientRightsException

    @property
    def auth_roles(self) -> t.List[Role]:
        # only filter roles that are defined in the ROLES dict
        return [
            self.build_role(role)
            for role in self.roles
            if role.split("/")[0] in ROLES
        ]

    @property
    def user_id(self) -> UUID4:
        return parse_obj_as(UUID4, self.sub.split(":")[-1])

    def assert_access(self, permission: t.Union[str, t.Sequence[str]]) -> None:
        if isinstance(permission, str):
            permission = [permission]

        for p in permission:
            self._is_allowed(
                p,
                (
                    self.build_role(role)
                    for role in self.roles
                    if role.split("/")[0] in ROLES
                ),
            )



import typing as t


def parse_permission_args(resource_str: str) -> t.Dict[str, str]:
    """
    Parse permission arguments from a resource string.

    Args:
        resource_str: String containing resource and optional permission args
            in the format: resource/arg1,arg2,{key1=val1,key2=val2}

    Returns:
        Dictionary of permission arguments
    """
    # Check for multiple argument blocks
    if resource_str.count("{") > 1:
        raise ValueError(
            f"Invalid permission string '{resource_str}'. "
            "Multiple argument blocks found"
        )
    if ",{" not in resource_str:
        return {}

    args_str = resource_str.split(",{", 1)[1]
    if not args_str.endswith("}"):
        raise ValueError(
            f"Invalid permission string '{resource_str}' ",
            "Missing closing brace",
        )
    args_str = args_str[:-1]  # Remove closing brace

    args = {}
    if args_str:
        for arg in args_str.split(","):
            if "=" not in arg:
                raise ValueError(
                    f"Invalid argument format in '{resource_str}'. "
                    "Expected key=value"
                )
            key, value = arg.split("=")
            args[key.strip()] = value.strip()

    return args


def parse_body(body: t.List[str]) -> t.List[t.Dict[str, t.Any]]:
    """Parse the body of a role definition."""
    clauses = []
    for clause in body:
        if ":" in clause:
            # This is a permission
            actions_str, resource_str = clause.split(":")

            # Extract permission arguments if present
            perm_args = parse_permission_args(resource_str)
            # Clean up resource string by removing argument block if present
            resource_str = resource_str.split("{")[0]

            resource_name, _, _ = resource_str.partition("/")
            actions = set(actions_str.split("+"))
            clauses.append(
                {
                    "type": "permission",
                    "actions": actions,
                    "resource_name": resource_name,
                    "args": perm_args,
                }
            )
        else:
            # This is a sub-role
            clauses.append({"type": "sub_role", "sub_role_name": clause})
    return clauses


def parse_permission(clause: str) -> t.Dict[str, t.Any]:
    """Parse a permission clause into its components.

    Format: action:resource/arg1,arg2,arg3,{key1=val1,key2=val2}

    Optional curly braces contain permission arg key-value pairs.
    A comma before the curly braces is required if args are present.
    """
    action_str, _, resource_str = clause.partition(":")
    actions = action_str.split("+") if action_str else []

    # Extract permission arguments if present
    perm_args = parse_permission_args(resource_str)
    # Clean up resource string by removing argument block if present
    resource_str = resource_str.split(",{")[0]

    resource_name, _, args_str = resource_str.partition("/")
    resource_args = args_str.split(",") if args_str else []

    return {
        "type": "permission",
        "actions": actions,
        "resource_name": resource_name,
        "resource_args": resource_args,
        "args": perm_args,
    }


def parse_sub_role(clause: str) -> t.Dict[str, t.Any]:
    sub_role, sep, sub_role_args_str = clause.partition("/")
    sub_role_args = sub_role_args_str.split(",") if sub_role_args_str else []
    return {
        "type": "role",
        "sub_role_name": sub_role,
        "sub_role_args": sub_role_args,
    }
